(** Minimalist text tetris implementation. *)

open Unix

let height = 15
let width = 11
let step_time = 1.0
let repeat_interval = 0.1
let score = ref 0

let board = Array.make_matrix height width ' '

let pieces =
  [
   [|[|' ';'*';' '|];
     [|'*';'*';'*'|]|];

   [|[|' ';'*';'*'|];
     [|'*';'*';' '|]|];

   [|[|'*';'*';' '|];
     [|' ';'*';'*'|]|];

   [|[|'*';' ';' '|];
     [|'*';'*';'*'|]|];

   [|[|' ';' ';'*'|];
     [|'*';'*';'*'|]|];

   [|[|'*';'*'|];
     [|'*';'*'|]|];

   [|[|'*';'*';'*';'*'|]|]
  ]

(** Prints the score and the board. *)
let print_board () =
  Printf.printf "Score: %d\n" !score;
  Printf.printf "+%s+\n" (String.make width '-');
  for i = 0 to height - 1 do
    Printf.printf "%c" '|';
    for j = 0 to width - 1 do
      Printf.printf "%c" board.(i).(j)
    done;
    Printf.printf "|\n";
  done;
  Printf.printf "+%s+\n" (String.make width '-');
  flush Pervasives.stdout

(** Uses [transform] to set the places on the board under [piece] put at
    [(y, x)]. [(y, x)] must be such that the piece is inside the board. Returns
    [true] iff a ['*'] on the board was set (a collision occurred). *)
let test_and_put piece (y, x) transform =
  let collision = ref false in
  for dy = 0 to Array.length piece - 1 do
    for dx = 0 to Array.length piece.(0) - 1 do
      if piece.(dy).(dx) == '*' then
      ( let v = board.(y + dy).(x + dx) in
        if v == '*' then collision := true;
        board.(y + dy).(x + dx) <- transform v )
    done
  done;
  !collision

let invert place = if place == '*' then ' ' else '*'
let set place = '*'
let put piece (y, x) = ignore (test_and_put piece (y, x) invert)
let remove = put

let rotate piece =
  let height = Array.length piece in
  let width = Array.length piece.(0) in
  let height', width' = width, height in
  let piece' = Array.make_matrix height' width' 'X' in
  for i = 0 to height' - 1 do
    for j = 0 to width' - 1 do
      piece'.(i).(j) <- piece.(j).(width - 1 - i)
    done
  done;
  piece'

(* There must not be any empty row or column on the edges of the piece. *)
let try_move piece cur_pos piece' (y', x') =
  let max_x' = x' + (Array.length piece'.(0)) - 1 in
  let max_y' = y' + (Array.length piece') - 1 in
  let inside_board = x' >= 0 && y' >= 0 && max_x' < width && max_y' < height in
  if not inside_board then false else
  ( remove piece cur_pos;
    if not (test_and_put piece' (y', x') invert) then true else
    ( remove piece' (y', x');
      put piece cur_pos;
      false ) )

let try_move' piece cur_pos piece' tentative_pos =
  if try_move piece cur_pos piece' tentative_pos then (piece', tentative_pos)
  else (piece, cur_pos)

let rec drop piece (y, x) =
  if try_move piece (y, x) piece (y + 1, x) then drop piece (y + 1, x)
  else (piece, (y, x))

let process_key k piece (y, x) =
  match k with
  | 'h' -> try_move' piece (y, x) piece (y, x - 1)
  | 'l' -> try_move' piece (y, x) piece (y, x + 1)
  | 'j' -> try_move' piece (y, x) piece (y + 1, x)
  | 'k' -> try_move' piece (y, x) (rotate piece) (y, x)
  | ' ' -> drop piece (y, x)
  | _   -> (piece, (y, x))

type read_key_t = Timeout | Key of char

let read_key timeout =
  if timeout <= 0.0 then Timeout else
  match select [stdin] [] [] timeout with
  | ([], _, _) -> Timeout
  | _          ->
      let buf = "x" in
      let _ = read stdin buf 0 1 in
      Key buf.[0]

let rec read_key_loop piece (y, x) deadline =
  match read_key (deadline -. gettimeofday ()) with
  | Timeout -> (piece, (y, x))
  | Key c   ->
      let (piece, (y, x)) = process_key c piece (y, x) in
      ignore (select [] [] [] repeat_interval);
      print_board ();
      read_key_loop piece (y, x) deadline
 
(** Returns the number of removed rows. *)
let opt_remove_rows () =
  let move_down y =
    for j = y - 1 downto 0 do
      board.(j + 1) <- Array.copy board.(j)
    done;
    board.(0) <- Array.create width ' '
  in let removed_row = ref 0 in
  for j = 0 to height - 1 do
    let f acc place = acc && place == '*' in
    if Array.fold_left f true board.(j) then
    ( incr removed_row;
      move_down j )
  done;
  !removed_row

(** Handles the fall of [piece] currently at position [(y, x)] (until it
    reaches the bottom). *)
let rec let_fall piece (y, x) =
  let deadline = gettimeofday () +. step_time in
  let (piece, (y, x)) = read_key_loop piece (y, x) deadline in
  if try_move piece (y, x) piece (y + 1, x) then
  ( print_board ();
    let_fall piece (y + 1, x) )
  else
    let removed_rows = opt_remove_rows () in
    score := !score + removed_rows

(** Selects a random piece and calls [let_fall]. Decides when to end the
    game. *)
let rec main_loop () =
  let piece = List.nth pieces ((Random.bits ()) mod (List.length pieces)) in
  let init_pos = (0, width / 2) in
  let collision = test_and_put piece init_pos set in
  print_board ();
  if collision then print_endline "Over"
  else
  ( let_fall piece init_pos;
    main_loop () )

let _ =
  Random.self_init ();
  let term_io = tcgetattr stdin in
  tcsetattr stdin TCSANOW { term_io with c_echo = false; c_icanon = false };
  main_loop ()
